void test_dot (void);
void test_nrm2 (void);
void test_asum (void);
void test_amax (void);
void test_axpy (void);
void test_copy (void);
void test_swap (void);
void test_scal (void);
void test_rotg (void);
void test_rot (void);
void test_rotmg (void);
void test_rotm (void);
void test_gemv (void);
void test_gbmv (void);
void test_trmv (void);
void test_tbmv (void);
void test_tpmv (void);
void test_symv (void);
void test_hemv (void);
void test_hbmv (void);
void test_sbmv (void);
void test_hpmv (void);
void test_spmv (void);
void test_trsv (void);
void test_tbsv (void);
void test_tpsv (void);
void test_ger (void);
void test_syr (void);
void test_her (void);
void test_hpr (void);
void test_spr (void);
void test_syr2 (void);
void test_spr2 (void);
void test_her2 (void);
void test_hpr2 (void);
void test_gemm (void);
void test_symm (void);
void test_hemm (void);
void test_syrk (void);
void test_herk (void);
void test_syr2k (void);
void test_her2k (void);
void test_trmm (void);
void test_trsm (void);
