  test_dot ();
  test_nrm2 ();
  test_asum ();
  test_amax ();
  test_axpy ();
  test_copy ();
  test_swap ();
  test_scal ();
  test_rotg ();
  test_rot ();
  test_rotmg ();
  test_rotm ();
  test_gemv ();
  test_gbmv ();
  test_trmv ();
  test_tbmv ();
  test_tpmv ();
  test_symv ();
  test_hemv ();
  test_hbmv ();
  test_sbmv ();
  test_hpmv ();
  test_spmv ();
  test_trsv ();
  test_tbsv ();
  test_tpsv ();
  test_ger ();
  test_syr ();
  test_her ();
  test_hpr ();
  test_spr ();
  test_syr2 ();
  test_spr2 ();
  test_her2 ();
  test_hpr2 ();
  test_gemm ();
  test_symm ();
  test_hemm ();
  test_syrk ();
  test_herk ();
  test_syr2k ();
  test_her2k ();
  test_trmm ();
  test_trsm ();
